from __future__ import annotations


def assert_equal(x, y):
    assert x == y


def assert_not_equal(x, y):
    assert x != y


def assert_true(x):
    assert x


def assert_false(x):
    assert not x
